#! /bin/bash
# backup.sh -- automated backup script
# by allen brunson  december 26 2017

# this software is provided under the MIT license:
# https://opensource.org/licenses/MIT

# script variables

BACKUPDIR="${HOME}/Library/Logs/Backup"
CARBONCOPYAPP="/Applications/Utilities/Carbon Copy Cloner.app/Contents/MacOS/ccc"
CARBONCOPYTASK="Backup Task"

# input unix start time, output human-readable elapsed time

function elapsedTime()
{
    local strt="$1"
    local stop=$(unixTime)
    local totl=$(($stop - $strt))
    local hour=$(($totl / 3600))
    local mins=$((($totl % 3600) / 60))
    local secs=$(($totl % 60))
    printf "elapsed time: %02d:%02d:%02d\n" $hour $mins $secs
}

# filename to use for the current backup log

function filenameMain()
{
    date "+${BACKUPDIR}/Backup-%Y-%m-%d-%H-%M.log"
}

# filename to use for one backup session

function filenameSession()
{
    local name="$1"
    echo "${BACKUPDIR}/session${name}.txt"
}

# filename to use for the in-progress backup log

function filenameInProgress()
{
    echo "${BACKUPDIR}/Current.log"
}

# command-line help

function help()
{
    local name=$(basename $0)

    echo ""
    echo "${name} -- automated backup script"
    echo "version 1.0  by allen brunson  allenbrunson.com"
    echo ""
    echo "usage: ${name} [command]"
    echo ""
    echo "backup         perform backup operation"
    echo "dryrun         fake backup operation, log only"
    echo "removeoldlogs  remove stale backup logs"
    echo "showlog        display most recent backup log"
    echo ""
}

# source the local profile, if it has not yet been run

function localProfile()
{
    if [[ -f ~/.profile && -z "${EDITOR}" ]]; then
        source ~/.profile
    fi
}

# create a backup log from all session logs

function logCreate()
{
    local fnam="$1"

    if [[ -z "${fnam}" ]]; then
        echo "script error"
        exit 5
    fi

    rm -f ${fnam}
    touch ${fnam}

    for next in ${BACKUPDIR}/*.txt; do
        cat ${next} >> ${fnam}
        echo "" >> ${fnam}
    done
}

# returns a timestamp two weeks in the past

function logOldestAge()
{
    local curr=$(unixTime)
    local lday=$((60 * 60 * 24))
    local offs=$((2 * 7 * $lday))
    echo $(($curr - $offs))
}

# remove outdated backup logs

function logRemoveOldLogs()
{
    local fage=$((0))
    local mage=$(logOldestAge)

    for fnam in ${BACKUPDIR}/*; do
        fage=$(stat -n -f "%m" ${fnam})
        if [[ $fage -le $mage ]]; then
            rm -f ${fnam}
        fi
    done
}

# remove current session logs

function logRemoveSessionLogs
{
    local fil1="${BACKUPDIR}/*.txt"
    local fil2=$(filenameInProgress)
    rm -f ${fil1}
    rm -f ${fil2}
}

# display the most recent backup log

function logShow()
{
    local fnam=""
    local test=$(find ${BACKUPDIR} -name "*.txt")

    # if a backup is in progress, create an in-progress backup log

    if [[ -nz "${test}" ]]; then
        logCreate $(filenameInProgress)
    fi

    # loop to find the most recent backup log

    for next in ${BACKUPDIR}/*.log; do
        fnam=${next}
    done

    # if no log files exist, exit

    if [[ -z "${fnam}" ]]; then
        echo "no files found"
        exit 1
    fi

    # display the most recent log file

    open ${fnam}
    exit 0
}

# log setup at the beginning of a backup

function logStart()
{
    local mode="$1"
    mkdir -p ${BACKUPDIR}
    logRemoveOldLogs
    logRemoveSessionLogs
}

# finalize log state at the end of a backup

function logStop()
{
    local name="$1"
    local mode="$2"
    logCreate "${name}"
    logRemoveSessionLogs
}

# script execution starts here

function main()
{
    local cmnd="$1"

    localProfile

    if [[ -z "${cmnd}" ]]; then
        help $*
        exit 1
    fi

    if [[ "${cmnd}" == "backup" ]]; then
        mainBack "live"
        exit 0
    fi

    if [[ "${cmnd}" == "dryrun" ]]; then
        mainBack "fake"
        exit 0
    fi

    if [[ "${cmnd}" == "removeoldlogs" ]]; then
        echo "removing stale logs"
        logRemoveOldLogs
        exit 0
    fi

    if [[ "${cmnd}" == "showlog" ]]; then
        logShow
        exit 0
    fi

    echo "unrecognized command: '${cmnd}'"
    help $*
    exit 1
}

function mainBack()
{
    local bfil=$(filenameMain)
    local mode="$1"

    # start subsystems

    sleepStart "${mode}"
    logStart "${mode}"

    # server sync has to run to completion before we continue

    # mainServer "${mode}"

    # start backup sessions and wait for them to finish

    mainBackBlaze "${mode}" &
    mainCarbonCopy "${mode}" &
    mainTimeMachine "${mode}" &
    wait
    sleep 2

    # finalize subsystems

    logStop "${bfil}" "${mode}"
    sleepStop "${bfil}" "${mode}"
}

# perform backblaze backup

function mainBackBlaze()
{
    local fnam=$(filenameSession backblaze)
    local mode="$1"
    local strt=$(unixTime)

    sessionStart "backblaze" 2>&1 1>> ${fnam}

    if [[ "${mode}" == "live" ]]; then
        /Library/Backblaze.bzpkg/bztransmit \
         -forcefullfilescan_backup_wait_for_completion \
         2>&1 1>> ${fnam}
    else
        echo "fake backblaze backup" 2>&1 1>> ${fnam}
    fi

    sessionStop ${strt} 2>&1 1>> ${fnam}
}

# perform carbon copy backup

function mainCarbonCopy()
{
    local buff=""
    local fnam=$(filenameSession carboncopycloner)
    local mode="$1"
    local strt=$(unixTime)

    # start the session log
    sessionStart "carbon copy cloner" 2>&1 1>> ${fnam}

    # perform the backup
    if [[ "${mode}" == "live" ]]; then
        "${CARBONCOPYAPP}" -s"${CARBONCOPYTASK}" -w 2>&1 1>> /dev/null
    fi

    # backup summary to buff
    buff=$("${CARBONCOPYAPP}" -h)

    # add most recent summary line to the session log
    echo "${buff}" | head -n 1 2>&1 1>> ${fnam}

    # stop the session log
    sessionStop ${strt} 2>&1 1>> ${fnam}
}

# perform server backup

function mainServer()
{
    local fnam=$(filenameSession serverbackup)
    local mode="$1"
    local strt=$(unixTime)

    sessionStart "server backup" 2>&1 1>> ${fnam}

    if [[ "${mode}" == "live" ]]; then
        echo "live server backup" 2>&1 1>> ${fnam}
    else
        echo "fake server backup" 2>&1 1>> ${fnam}
    fi

    sessionStop "${strt}" 2>&1 1>> ${fnam}
}

# perform time machine backup

function mainTimeMachine()
{
    local fnam=$(filenameSession timemachine)
    local mode="$1"
    local strt=$(unixTime)

    sessionStart "time machine" 2>&1 1>> ${fnam}

    if [[ "${mode}" == "live" ]]; then
        tmutil startbackup --block 2>&1 1>> ${fnam}
    else
        echo "fake time machine backup" 2>&1 1>> ${fnam}
    fi

    sessionStop "${strt}" 2>&1 1>> ${fnam}
}

# start backup session with a given name

function sessionStart()
{
    local name=$@
    date "+${name} backup start: %B %d %Y %I:%M%p"
}

# stop backup session with a given start time

function sessionStop()
{
    local strt="$1"
    elapsedTime ${strt}
}

# current display state. return values:
# 4: display is awake
# 3: display is dimmed, about to fall asleep
# 2, 1, 0: display is asleep

function sleepDisplayState()
{
    local disp=$(ioreg -n IODisplayWrangler -r IODisplayWrangler -w 0)
    local cval=$(echo "${disp}" | grep -o "\"CurrentPowerState\"=[0-9]")
    local pval=$(echo "${cval}" | grep -o "[0-9]")
    echo "${pval}"
}

function sleepStart()
{
    local mode="$1"

    if [[ "${mode}" == "live" ]]; then
        # put the display to sleep, but leave the computer running
        pmset displaysleepnow > /dev/null
    fi
}

function sleepStop()
{
    local fnam="$1"
    local mode="$2"
    local rval=$(sleepDisplayState)

    if [[ $rval -ge 4 ]]; then
        echo "display is awake, exiting" >> ${fnam}
    else
        echo "putting the computer to sleep" >> ${fnam}
        pmset sleepnow > /dev/null
    fi
}

# return current time as a timestamp

function unixTime()
{
    date "+%s"
}

# script execution starts here

main $*
